`timescale 1ns / 1ps

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------
module AD7961
    (
        input           m_clk_i,                    // 100 MHz Clock, used for tiing
        input           fast_clk_i,                 // Maximum 300 MHz Clock, used for serial transfer
        input           reset_n_i,                  // Reset signal, active low
        input   [ 3:0]  en_i,                       // Enable pins input
        input           d_pos_i,                    // Data In, Positive Pair
        input           d_neg_i,                    // Data In, Negative Pair
        input           dco_pos_i,                  // Echoed Clock In, Positive Pair
        input           dco_neg_i,                  // Echoed Clock In, Negative Pair
        output  [ 3:0]  en_o,                       // Enable pins output
        output          cnv_pos_o,                  // Convert Out, Positive Pair
        output          cnv_neg_o,                  // Convert Out, Negative Pair
        output          clk_pos_o,                  // Clock Out, Positive Pair
        output          clk_neg_o,                  // Clock Out, Negative Pair
        output          data_rd_rdy_o,              // Signals that new data is available
        output  [15:0]  data_o                      // Read Data
    );

//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------
// FPGA Clock Frequency
parameter real          FPGA_CLOCK_FREQ = 100;

// Conversion signal generation
parameter real          TCYC                    = 0.200;
parameter       [31:0]  ADC_CYC_CNT             = FPGA_CLOCK_FREQ * TCYC - 1;

// Serial Interface
parameter               SERIAL_IDLE_STATE       = 3'b001;
parameter               SERIAL_READ_STATE       = 3'b010;
parameter               SERIAL_DONE_STATE       = 3'b100; 
 
//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------ 
// Conversion signal generation 
reg [31:0]  adc_tcyc_cnt;
wire         cnv_s; 

// TMSB Counter
wire        tmsb_done_s;

// Serial Interface
reg [ 2:0]  serial_present_state;
reg [ 2:0]  serial_next_state;
reg [ 2:0]  serial_present_state_d1;
reg [31:0]  sclk_cnt;
reg [18:0]  serial_buffer;
wire        buffer_reset_s;
reg         serial_read_done_s;

//------------------------------------------------------------------------------
//----------- Wires Declarations -----------------------------------------------
//------------------------------------------------------------------------------
// Serial Interface
wire        clk_s;
wire        sclk_s;
wire        sdi_s; 
wire [18:0] data_o_s;
wire        fast_clk_s;

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------
assign clk_s = ((serial_present_state_d1 == SERIAL_READ_STATE)&&(serial_present_state != SERIAL_DONE_STATE)) ? 1'b1 : 1'b0; 
assign en_o = en_i;
assign data_o_s = (serial_present_state == SERIAL_DONE_STATE) ? serial_buffer : data_o_s;
//assign data_o = data_o_s[18:1];
assign data_o = data_o_s[15:0];
assign data_rd_rdy_o = serial_read_done_s & (adc_tcyc_cnt < 32'd4); 

ODDR2 #(
    .DDR_ALIGNMENT("NONE"), // Sets output alignment to "NONE", "C0" or "C1" 
    .INIT(1'b0),    // Sets initial state of the Q output to 1'b0 or 1'b1
    .SRTYPE("SYNC") // Specifies "SYNC" or "ASYNC" set/reset
) ODDR2_inst (
    .Q(fast_clk_s),   // 1-bit DDR output data
    .C0(fast_clk_i), // 1-bit clock input
    .C1(~fast_clk_i),  // 1-bit clock input
    .CE(clk_s), // 1-bit clock enable input
    .D0(1'b1), // 1-bit data input (associated with C0)
    .D1(1'b0), // 1-bit data input (associated with C1)
    .R(1'b0),   // 1-bit reset input
    .S(1'b0)    // 1-bit set input
);

// Data In LVDS -> Single
IBUFDS 
    #( 
        .DIFF_TERM("TRUE"),       // Differential Termination
        .IOSTANDARD("LVDS_25")     // Specify the input I/O standard
    ) 
    Data_In_IBUFDS 
    (
        .O(sdi_s),                  // Buffer output
        .I(d_pos_i),                // Diff_p buffer input (connect directly to top-level port)
        .IB(d_neg_i)                // Diff_n buffer input (connect directly to top-level port)
    );

// Conversion Out Single -> LVDS
OBUFDS 
    #(
        .IOSTANDARD("LVDS_25")      // Specify the output I/O standard
    ) 
    Cnv_Out_OBUFDS 
    (
        .O(cnv_pos_o),              // Diff_p output (connect directly to top-level port)
        .OB(cnv_neg_o),             // Diff_n output (connect directly to top-level port)
        .I(cnv_s)                   // Buffer input 
    );

// Clock Out Single -> LVDS    
OBUFDS 
    #(
        .IOSTANDARD("LVDS_25")      // Specify the output I/O standard
    ) 
    Clock_Out_OBUFDS 
    (
        .O(clk_pos_o),              // Diff_p output (connect directly to top-level port)
        .OB(clk_neg_o),             // Diff_n output (connect directly to top-level port)
        .I(fast_clk_s)              // Buffer input 
    );    

// Update conversion timing counters
always @(posedge m_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_tcyc_cnt <= ADC_CYC_CNT;
    end
    else
    begin
        if(adc_tcyc_cnt != 32'd0)
        begin
            adc_tcyc_cnt <= adc_tcyc_cnt - 32'd1;
        end
        else
        begin
            adc_tcyc_cnt <= ADC_CYC_CNT;
        end
    end
end 

assign cnv_s = (adc_tcyc_cnt > 32'd14) ? 1'b1 : 1'b0;
assign tmsb_done_s = (adc_tcyc_cnt == 32'd19) ? 1'b1 : 1'b0;


// Register States
always @(posedge fast_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        serial_present_state <= SERIAL_IDLE_STATE;
    end
    else
    begin
        serial_present_state <= serial_next_state;
        serial_present_state_d1 <= serial_present_state;
    end
end

// State switch logic
always @(tmsb_done_s, sclk_cnt, serial_present_state) 
begin
    serial_next_state <= serial_present_state;
    case(serial_present_state)
        SERIAL_IDLE_STATE:
            begin
                if(tmsb_done_s == 1'b1)
                begin
                    serial_next_state <= SERIAL_READ_STATE;
                end
            end
        SERIAL_READ_STATE:
            begin
                if(sclk_cnt == 32'd0) 
                begin
                    serial_next_state <= SERIAL_DONE_STATE;
                end
            end
        SERIAL_DONE_STATE:
            begin
                serial_next_state <= SERIAL_IDLE_STATE;
            end
        default:
            begin
                serial_next_state <= SERIAL_IDLE_STATE;
            end
    endcase
end

// State output logic
always @(posedge fast_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        serial_read_done_s <= 1'b0;
    end
    else
    begin
        case(serial_present_state)
            SERIAL_IDLE_STATE:
                begin
                    serial_read_done_s <= 1'b1;
                end
            SERIAL_READ_STATE:
                begin
                    serial_read_done_s <= 1'b0;
                end
            SERIAL_DONE_STATE:
                begin
                    serial_read_done_s <= 1'b1;
                end
            default: 
                begin   
                    serial_read_done_s <= 1'b0;
                end
        endcase
    end
end

assign buffer_reset_s = (adc_tcyc_cnt == 32'd1) ? 1'b1 : 1'b0;
 
always @(posedge fast_clk_i or posedge buffer_reset_s)
begin
    if(buffer_reset_s == 1'b1)
    begin 
        serial_buffer <= 21'd0;
        //sclk_cnt <= 32'd23;
        sclk_cnt <= 32'd20;
    end
    else if((serial_present_state == SERIAL_READ_STATE)&&(sclk_cnt > 32'd0)&&(clk_s == 1'b1))
    begin
        sclk_cnt <= sclk_cnt - 32'd1;
        serial_buffer <= {serial_buffer[17:0], sdi_s};
    end
end

endmodule
